//The following code examples are adapted from the book
//The C++ Standard Library - A Tutorial and Reference, 2nd Edition
//by Nicolai M.Josuttis, Addison Wesley Longman, 2012
//Copyright  2012 by Pearson Education, Inc.and Nicolai M.Josuttis

#include <vector>
#include <algorithm>
#include <iostream>
using namespace std;

// function that prints the passed argument
void print(int elem)
{
	cout << elem << ' ';
}


// simple function object that prints the passed argument
class PrintInt {
public:
	void operator() (int elem) const {
		cout << elem << ' ';
	}
};


// function object that adds the value with which it is initialized
class AddValue 
{
	private:
		int theValue;    // the value to add
	public:
		// constructor initializes the value to add
		AddValue(int v) : theValue(v) 
		{
		}

		// the "function call" for the element adds the value
		void operator() (int& elem) const 
		{
			elem += theValue;
		}
};


// function object to process the mean value
class MeanValue 
{
	private:
		long num;    // number of elements
		long sum;    // sum of all element values
	public:
		// constructor
		MeanValue() : num(0), sum(0) 
		{
		}

		// "function call"
		// - process one more element of the sequence
		void operator() (int elem) 
		{
			++num;          // increment count
			sum += elem;    // add value
		}

		// return mean value
		double value() 
		{
			return static_cast<double>(sum) / static_cast<double>(num);
		}
};


int main()
{
	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #1: function as an arugument for function call" << endl
		<< "**************************************************" << endl;

	vector<int> coll1 = { 1, 2, 3, 4, 5 };

	// print all elements
	for_each(coll1.begin(), coll1.end(),  // range
		print);                      // operation
	cout << endl;


	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #2:  function object as an arugument for function call" << endl
		<< "**************************************************" << endl;

	vector<int> coll2 = { 5, 4, 3, 2, 1 };
	// print all elements
	for_each(coll2.begin(), coll2.end(),  // range
		PrintInt());                 // operation
	cout << endl;
	


	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #3: function object with internal state" << endl
		<< "**************************************************" << endl;

	vector<int> coll3 = { 2, 4, 6, 8, 10 };
	// add value 10 to each element
	for_each(coll3.begin(), coll3.end(),    // range
		AddValue(10));               // operation


	for_each(coll3.begin(), coll3.end(),  // range
		PrintInt());                 // operation
	cout << endl;


	// add value of first element to each element
	for_each(coll3.begin(), coll3.end(),    // range
		AddValue(*coll3.begin()));    // operation

	for_each(coll3.begin(), coll3.end(),  // range
		PrintInt());                 // operation
	cout << endl;




	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #4: function object with internal state" << endl
		<< "**************************************************" << endl;

	vector<int> coll4 = { 2, 4, 6, 8, 10 };
	MeanValue mv;

	mv= for_each(coll4.begin(), coll4.end(),  // range
		MeanValue());              // operation
	cout << "mean value of coll4: " << mv.value() << endl;


	vector<int> coll5 = { 1, 3, 5, 7, 9 };
	// process and print mean value
	mv = for_each(coll5.begin(), coll5.end(),  // range
		MeanValue());              // operation
	cout << "mean value of coll5: " << mv.value() << endl;


	// Wait for the user input before ending the program
	char inputCharacter;    // Declare a variable for storing a character input from the keyboard
	cin >> inputCharacter; // Wait to read in a character input from the keyboard


	//Finish the program and return the control to the operating system.
	return 0;


}





