//The following code examples are adapted from the book
//The C++ Standard Library - A Tutorial and Reference, 2nd Edition
//by Nicolai M.Josuttis, Addison Wesley Longman, 2012
//Copyright  2012 by Pearson Education, Inc.and Nicolai M.Josuttis

#include<functional>
#include <vector>
#include <algorithm>
#include <iostream>
using namespace std;

void capture1() //Capture by value
{
	int v = 10;
	auto f = [v] { return v*v; };

	v = 0;
	cout << f() << endl;

}


void capture2() //Capture by reference
{
	int v = 10;
	auto f = [&v] { return v*v; };

	v = 0;
	cout << f() << endl;

}


function<int(int, int)> returnLambda1()
{
	return 
	[](int x, int y) 
	{
		return x*y;
	};
}

function<int(int)> returnLambda2(int exponent)
{
	return
		[exponent](int base)
	{
		int result=1;
		for (int i = 0; i < exponent; i++)
			result *= base;

		return result;
	};
}

function<int(int)> returnLambdaX(int exponent)
{
	return
		[& exponent](int base) //Not quite right using [& exponent] here
	{
		int result = 1;
		for (int i = 0; i < exponent; i++)
			result *= base;

		return result;
	};
}




int main()
{
	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #1: Capture by value" << endl
		<< "**************************************************" << endl;

	capture1();

	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #2: Capture by reference" << endl
		<< "**************************************************" << endl;

	capture2();

	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #3: lambda expressions, capture by reference, explicit" << endl
		<< "**************************************************" << endl;

	vector<int> coll1 = { 1, 2, 3, 4, 5 };

	// process and print mean value
	long sum1 = 0;
	for_each(coll1.begin(), coll1.end(),  // range
		[&sum1](int elem) {
		sum1 += elem;
	});
	cout << "Sum: " << sum1 << endl;


	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #4: lambda expressions, capture by reference, implicit" << endl
		<< "**************************************************" << endl;

	vector<int> coll2 = { 1, 2, 3, 4, 5 };

	// process and print mean value
	long sum2 = 0;
	for_each(coll2.begin(), coll2.end(),  // range
		[&](int elem) {
		sum2 += elem;
	});
	cout << "Sum: " << sum2 << endl;

	//*************************************************
	//What would happen if you fo the follow? Why?
	//*************************************************
	//cout << endl << endl
	//	<< "**************************************************" << endl
	//	<< "Experiment #4XX: lambda expressions, capture by value" << endl
	//	<< "**************************************************" << endl;

	//vector<int> coll2 = { 1, 2, 3, 4, 5 };

	//// process and print mean value
	//long sum2 = 0;
	//for_each(coll2.begin(), coll2.end(),  // range
	//	[sum2](int elem) {
	//	sum2 += elem;
	//});
	//cout << "Sum: " << sum2 << endl;



	cout << endl << endl
		<< endl << endl
		<< "**************************************************" << endl
		<< "Experiment #5: lambda expressions as function objects" << endl
		<< "**************************************************" << endl;

	auto lf1 = returnLambda1();
	cout << lf1(3, 4) << endl;
	cout << lf1(6, 9) << endl;

	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #6: lambda expressions as function objects" << endl
		<< "**************************************************" << endl;

	auto lf2 = returnLambda2(2);
	cout << lf2(1) << endl;
	cout << lf2(2) << endl;
	cout << lf2(3) << endl;
	cout << lf2(4) << endl;


	cout << endl << endl
		<< "**************************************************" << endl
		<< "Experiment #7: lambda expressions as function objects" << endl
		<< "**************************************************" << endl;
	auto lf3 = returnLambda2(3);
	cout << lf3(1) << endl;
	cout << lf3(2) << endl;
	cout << lf3(3) << endl;
	cout << lf3(4) << endl;

	//*************************************************
	//What would happen if you fo the follow? Why?
	//*************************************************
	//cout << endl << endl
	//	<< "**************************************************" << endl
	//	<< "Experiment #7XXX: lambda expressions as function objects" << endl
	//	<< "**************************************************" << endl;
	//auto lf4 = returnLambdaX(3);
	//cout << lf4(1) << endl;
	//cout << lf4(2) << endl;
	//cout << lf4(3) << endl;
	//cout << lf4(4) << endl;

	// Wait for the user input before ending the program
	char inputCharacter;    // Declare a variable for storing a character input from the keyboard
	cin >> inputCharacter; // Wait to read in a character input from the keyboard


	//Finish the program and return the control to the operating system.
	return 0;


}





